import { CallExpression } from '@babel/types';
import { EnvContext as EnvContext_2 } from '@typed-macro/core';
import { File as File_2 } from '@babel/types';
import { FSWatcher } from 'chokidar';
import generate from '@babel/generator';
import { Host } from '@typed-macro/shared';
import { ImportDeclaration } from '@babel/types';
import { Node as Node_2 } from '@babel/types';
import { NodePath } from '@babel/traverse';
import { PackageManager } from '@typed-macro/shared';
import { parse } from '@babel/parser';
import { parseExpression } from '@babel/parser';
import { ParserPlugin } from '@babel/parser';
import { Program } from '@babel/types';
import { State as State_2 } from '@typed-macro/core';
import template from '@babel/template';
import { Thenable } from '@typed-macro/shared';
import { TransformResult } from '@typed-macro/shared';
import traverse from '@babel/traverse';
import * as types from '@babel/types';

export declare type Babel = {
    /**
     * From \@babel/template.
     * Help generate nodes from code string.
     *
     * e.g.
     * ```typescript
     * template.statement.ast('const a = 1') // generate a statement node
     * ```
     */
    template: typeof template;
    /**
     * From \@babel/traverse.
     * Help traverse from a node.
     *
     * e.g.
     * ```typescript
     * traverse(someNode, {
     *   CallExpression(path) {
     *     ...
     *   }
     * })
     * ```
     */
    traverse: typeof traverse;
    /**
     * From \@babel/types.
     * Help check node type and create nodes of specific type.
     *
     * e.g.
     * ```typescript
     *  types.isIdentifier(someNode) // check is identifier
     *  types.identifier('hello') // create identifier node
     * ```
     */
    types: typeof types;
    /**
     * From \@babel/parser.
     * Help parse code string to ast.
     *
     * e.g.
     * ```typescript
     *  parse('const a = 1')
     * ```
     */
    parse: typeof parse;
    /**
     * From \@babel/parser.
     * Help parse code string to expression.
     *
     * e.g.
     * ```typescript
     *  parseExpression('1 + 2')
     * ```
     */
    parseExpression: typeof parseExpression;
    /**
     * From \@babel/generator.
     * Help generate code string from ast.
     *
     * e.g.
     * ```typescript
     *  generate(someNode)
     * ```
     */
    generate: typeof generate;
};

/**
 * Define a macro.
 *
 * For example,
 * ```typescript
 * defineMacro('someMacro')
 *   .withCustomType('export type SomeType = string')
 *   .withSignature('(): SomeType', 'some comments')
 *   .withHandler(({ path }) => {
 *     path.remove()
 *   })
 * ```
 *
 * @param name - the name of the macro, should be a valid identifier name.
 */
export declare function defineMacro(name: string): Omit<MacroBuilder, 'withHandler'>;

/**
 * Define a macro provider.
 */
export declare function defineMacroProvider(raw: MacroProvider): MacroProvider;

export declare type EnvContext = {
    /**
     * The bundler/dev server in which macros/providers run.
     * @example 'vite'
     */
    host: Host;
    /**
     * The package manager of the project to which macros/providers are applied.
     * @example 'yarn'
     */
    packageManager: PackageManager;
    /**
     * The project path.
     *
     * If the current project is a subproject (for example, in a workspace),
     * the projectPath will be like
     * `[currentProjectDirectory, parentProjectDirectory, ..., rootProjectDirectory]`.
     */
    projectPath: string[];
    /**
     * Is in dev mode.
     *
     * Different {@link EnvContext.host} can have different judgment strategy.
     */
    dev: boolean;
    /**
     * Is in SSR mode.
     *
     * Different {@link EnvContext.host} can have different judgment strategy.
     */
    ssr: boolean;
    /**
     * An instance of chokidar's FSWatcher.
     *
     * It will be undefined in building process.
     */
    watcher?: FSWatcher;
    /**
     * Used to manage modules.
     *
     * It will be undefined in building process.
     */
    modules?: Modules;
};

/**
 * An Exportable is a virtual module holding macros **or** code.
 *
 * If both macros and code are set, the macros would be ignored.
 */
export declare type Exportable = ({
    /**
     * macros
     */
    macros: Macro[];
} | {
    /**
     * code in **Javascript**
     */
    code: string;
}) & {
    /**
     * Type definitions, will be written to d.ts.
     *
     * e.g.
     * ```typescript
     * { exports: { '@macros': { types: `type A = string` } } }
     * ```
     * will generate
     * ```typescript
     * declare module '@macros' {
     *   type A = string
     * }
     * ```
     */
    types?: string;
};

/**
 * ```typescript
 * import 'moduleName'
 * { moduleName: string }
 *
 * import defaultName from 'moduleName'
 * { defaultName: string; moduleName: string }
 *
 * import { exportName } from 'moduleName'
 * { exportName: string; moduleName: string }
 *
 * import { exportName as localName } from 'moduleName'
 * { localName: string; exportName: string; moduleName: string }
 *
 * import * as namespaceName from 'moduleName'
 * { namespaceName: string; moduleName: string }
 * ```
 */
export declare type ImportOption = {
    moduleName: string;
} | {
    defaultName: string;
    moduleName: string;
} | {
    localName: string;
    exportName: string;
    moduleName: string;
} | {
    exportName: string;
    moduleName: string;
} | {
    namespaceName: string;
    moduleName: string;
};

export declare type InstantiatedMacroProvider = {
    id: string;
    /**
     * Exports macros so macros can be imported and called,
     * the key will be the name of the virtual module.
     *
     * Also supports exporting source codes,
     * but a module cannot export both codes and macros.
     *
     * e.g.
     * ```typescript
     * const macroA = defineMacro('macroA')
     *   .withSignature('(...args: any[]): void')
     *   .withHandler(()=>...)
     *
     * {
     *   exports: {
     *     '@macros': { macros: [macroA] },
     *     '@helper': {
     *        code: `export const a = 1`,
     *        types: `export const a:number`
     *      }
     *    },
     *  }
     * ```
     * Then in some .js(x) or .ts(x) file you can
     * ```typescript
     * import { macroA } from '@macros'
     * import { a } from '@helper'
     * macroA(someArgs)
     * console.log(a)
     * ```
     */
    exports: ModularizedExportable;
    /**
     * Provider hooks.
     */
    hooks?: MacroProviderHooks;
    /**
     * Provider options.
     */
    options?: MacroProviderOptions;
};

export declare function isMacro(o: unknown): o is Macro;

export declare function isMacroProvider(o: unknown): o is MacroProvider;

export declare type Macro = {
    /* Excluded from this release type: __is_macro */
    /* Excluded from this release type: name */
    /* Excluded from this release type: handler */
    /* Excluded from this release type: types */
    /* Excluded from this release type: version */
};

export declare type MacroBuilder = {
    /**
     * Add custom type definitions.
     *
     * They will be rendered with macro signatures.
     *
     * @param typeDefinition like 'export type SomeType = string'
     */
    withCustomType(typeDefinition: string): Omit<MacroBuilder, 'withHandler'>;
    /**
     * Add signature of macro. one macro requires at least one signature.
     * @param signature - function signature, like '(s: string): void'
     * @param comment - comment for the signature, will be written to .d.ts
     * before the corresponding function signature.
     */
    withSignature(signature: string, comment?: string): Omit<MacroBuilder, 'withCustomType'>;
    /**
     * Set the transform handler and get the macro.
     *
     * @see MacroHandler
     */
    withHandler(handler: MacroHandler): Macro;
};

/* Excluded from this release type: MacroCall */

export declare type MacroContext = EnvContext_2 & MacroTransformContext;

/**
 * Async function can not be used as macro handler.
 *
 * If the handler is a normal function, the nested macros inside the current
 * call expression will be expanded automatically before calling the handler.
 *
 * If the handler is a generator function, you can:
 *  - yield node paths of import statements to collect macros from them,
 *    note that macros must be collected before used, or you can wait for the next traversal
 *    (the runtime collects imported macros before every traversal),
 *  - yield node paths of other types to actively expand macros inside them.
 *
 * e.g.
 * ```typescript
 * function* someMacro(ctx) {
 *   // do some thing
 *   yield ctx.args // expand macros inside arguments
 *   // do some thing
 *   yield someOtherNodePath
 *   // do some thing
 * }
 * ```
 */
export declare type MacroHandler = (
/**
 * Macro call context.
 */
ctx: MacroContext, 
/**
 * Babel tools to help operate node paths and nodes.
 */
babel: Readonly<Babel>, 
/**
 * Wrappers upon babel tools to simplify operations.
 */
helper: MacroHelper) => SuspendableTransform | void;

export declare type MacroHelper = {
    /**
     * Prepend import statements to program.
     *
     * If a same import already exists, it will not be re-imported.
     *
     * @param imports - an array of import options
     * @param program - node path of the target program for prepending import statements.
     * use the one currently being handled by default.
     * @return an array of node paths of the inserted import statements or the existed import statements,
     * in the same order of the input imports options.
     */
    prependImports(imports: ImportOption[], program?: NodePath<Program>): NodePath<ImportDeclaration>[];
    /**
     * Prepend an import statement to program.
     * @see MacroHelper.prependImports
     */
    prependImports(imports: ImportOption, program?: NodePath<Program>): NodePath<ImportDeclaration>;
    /**
     * Append import statements to the last import statement of the program.
     *
     * Note that the order of import statements may affect the order in which modules are executed.
     *
     * @see https://stackoverflow.com/questions/35551366/69822046#69822046
     * @see MacroHelper.prependImports
     */
    appendImports(imports: ImportOption[], program?: NodePath<Program>): NodePath<ImportDeclaration>[];
    /**
     * Append an import statement to the last import statement of the program.
     * @see MacroHelper.appendImports
     */
    appendImports(imports: ImportOption, program?: NodePath<Program>): NodePath<ImportDeclaration>;
    /**
     * Find import statements that exist already.
     *
     * Note that findImported() can not find imported macros.
     *
     * @param imports - import options
     * @param loose - match loosely, e.g:
     * `import { a as _a } from 'a'` will match both `{ moduleName: 'a' }` and `{ moduleName: 'a', exportName: 'a' }`.
     *  Defaults to false.
     * @param program - node path of the target program. use the one currently being handled by default.
     */
    findImported(imports: ImportOption[], loose?: boolean, program?: NodePath<Program>): (NodePath<ImportDeclaration> | undefined)[];
    /**
     * Find an import statement that exists already.
     *
     * @see MacroHelper.findImported
     */
    findImported(imports: ImportOption, loose?: boolean, program?: NodePath<Program>): NodePath<ImportDeclaration> | undefined;
    /**
     * Prepend nodes to the target program.
     *
     * @param nodes - an array of nodes
     * @param program - node path of the target program. use the one currently being handled by default.
     * @return an array of node paths of the inserted nodes
     */
    prependToBody(nodes: Node_2[], program?: NodePath<Program>): NodePath[];
    /**
     * Prepend a node to the target program.
     * @see MacroHelper.prependToBody
     */
    prependToBody(node: Node_2, program?: NodePath<Program>): NodePath;
    /**
     * Append any node to the target program.
     * @see MacroHelper.prependToBody
     */
    appendToBody(nodes: Node_2[], program?: NodePath<Program>): NodePath[];
    /**
     * Append a node to the target program.
     * @see MacroHelper.appendToBody
     */
    appendToBody(nodes: Node_2, program?: NodePath<Program>): NodePath;
    /**
     * Check whether there are unexpanded macros inside the nodes.
     * @param paths - the node path to be checked
     */
    containsMacros(paths: NodePath[]): boolean[];
    /**
     * Check whether there are unexpanded macros inside the node.
     * @see MacroHelper.containsMacros
     */
    containsMacros(path: NodePath): boolean;
};

export declare type MacroProvider = {
    /* Excluded from this release type: __is_macro_provider */
    /* Excluded from this release type: version */
} & (InstantiatedMacroProvider | ((env: EnvContext) => InstantiatedMacroProvider));

/**
 * Hooks Order
 * ```
 * onStart()
 * transform files loop:
 *   onFilter()
 *   if filter passed:
 *     beforeTransform()
 *     apply macros
 *     afterTransform()
 * onStop()
 * ```
 */
export declare type MacroProviderHooks = {
    /**
     * Called after initialized.
     */
    onStart?(): Thenable<void>;
    /**
     * Called when determining whether a file should be transformed.
     *
     * Note that an excluded file will never be passed to this hook.
     */
    onFilter?(id: string): Thenable<boolean>;
    /**
     * Called before transformation by macros.
     *
     * Can turn some non-js/non-ts file into JS code in this hook.
     */
    beforeTransform?(code: string, id: string): Thenable<TransformResult>;
    /**
     * Called after transformation by macros.
     */
    afterTransform?(code: string, id: string): Thenable<TransformResult>;
    /**
     * Called before dev server stops or bundler finishes building.
     */
    onStop?(): Thenable<void>;
};

export declare type MacroProviderOptions = {
    /**
     * Babel plugins to be applied during parsing.
     *
     * 'typescript' and 'jsx' are included by default and cannot be removed.
     *
     * Some plugins are enabled automatically by the latest babel parser
     * because they have been regarded as part of the language.
     *
     * @see https://babeljs.io/docs/en/babel-parser#plugins
     * @see https://babeljs.io/docs/en/babel-parser#latest-ecmascript-features
     */
    parserPlugins?: ParserPlugin[];
};

export declare type MacroTransformContext = {
    /**
     * The file path of the module currently being handled.
     */
    filepath: string;
    /**
     * The original source code of the module currently being handled.
     */
    code: string;
    /**
     * The arguments node paths of the call-macro expression currently being handled.
     */
    args: CallExpression['arguments'] extends Array<infer R> ? NodePath<R>[] : never;
    /**
     * The NodePath of the call-macro expression currently being handled.
     */
    path: NodePath<CallExpression>;
    /**
     * The AST being traversed.
     */
    ast: File_2;
    /**
     * Traversal state.
     *
     * In order to recursively expand all macros in a file,
     * the transformer will traverse the AST many times until all macros
     * are expanded.
     *
     * - {@link MacroTransformContext.state.traversal} is the state shared during one traversal, and
     * - {@link MacroTransformContext.state.transform} is the state shared during the whole transformation.
     */
    state: {
        /**
         * @see MacroContext.state
         */
        traversal: State_2;
        /**
         * @see MacroContext.state
         */
        transform: State_2;
    };
};

/**
 * Organize {@link Exportable}s.
 */
export declare type ModularizedExportable = {
    [moduleName: string]: Exportable;
};

/**
 * Manage modules, users can set tag of a module and then query/invalidate
 * modules by the tag.
 */
export declare type Modules = {
    /**
     * Get the tag of specific module.
     * @param id module filepath, see {@link MacroTransformContext.filepath}.
     */
    getTag(id: string): string | undefined;
    /**
     * Set tag for specific module.
     *
     * Note that one module can only have one tag, and the later tag override the earlier.
     *
     * @param id module filepath, see {@link MacroTransformContext.filepath}.
     * @param tag tag value
     */
    setTag(id: string, tag: string): void;
    /**
     * Unset tag for specific module.
     * @param id module filepath, see {@link MacroTransformContext.filepath}.
     */
    unsetTag(id: string): void;
    /**
     * Query modules having tags that match the pattern.
     *
     * @param pattern
     *   if string, a tag is matched when strictly equals to the pattern string;
     *   if RegExp, a tag is matched when pattern.test(tag) is true.
     */
    queryByTag(pattern: RegExp | string): string[];
    /**
     * Invalidate modules having tags that match the pattern.
     *
     * Use {@link Modules.queryByTag} to get the matched modules internally.
     */
    invalidateByTag(pattern: RegExp | string): string[];
};

export declare type State = {
    /**
     * Check whether value exists by key.
     */
    has<K>(key: K): boolean;
    /**
     * Get value by key.
     */
    get<T, K>(key: K): T;
    /**
     * Get value by key, or set value if not exists.
     */
    getOrSet<T, K>(key: K, or: T): T;
    /**
     * Set value by key.
     */
    set<T, K>(key: K, value: T): T;
    /**
     * Delete value by key.
     */
    delete<K>(key: K): void;
};

export declare type SuspendableTransform = Generator<YieldTask, any, any>;

export declare type YieldTask = NodePath | NodePath[] | undefined;

export { }
